# Copyright 2024 Mattia Giambirtone & All Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

## Implementation of negamax with a/b pruning
import board
import movegen
import eval
import transpositions


import std/math
import std/times
import std/atomics
import std/algorithm
import std/monotimes
import std/strformat


const NUM_KILLERS* = 2
const MAX_DEPTH* = 255


func computeLMRTable: array[MAX_DEPTH, array[218, int]] {.compileTime.} =
    ## Precomputes the table containing reduction offsets at compile
    ## time
    for i in 1..result.high():
        for j in 1..result[0].high():
            result[i][j] = round(0.8 + ln(i.float) * ln(j.float) * 0.4).int


const LMR_TABLE {.used.} = computeLMRTable()


type
    HistoryTable* = array[PieceColor.White..PieceColor.Black, array[Square(0)..Square(63), array[Square(0)..Square(63), Score]]]
    KillersTable* = array[MAX_DEPTH, array[NUM_KILLERS, Move]]
    SearchManager* = object
        ## A simple state storage
        ## for our search
        searchFlag: ptr Atomic[bool]
        stopFlag: ptr Atomic[bool]
        board: Chessboard
        bestRootScore: Score
        searchStart: MonoTime
        hardLimit: MonoTime
        softLimit: MonoTime
        nodeCount: uint64
        maxNodes: uint64
        searchMoves: seq[Move]
        transpositionTable: ptr TTable
        history: ptr HistoryTable
        killers: ptr KillersTable
        # We keep one extra entry so we don't need any special casing
        # inside the search function when constructing pv lines
        pvMoves: array[MAX_DEPTH + 1, array[MAX_DEPTH + 1, Move]]
        selectiveDepth: int


proc newSearchManager*(board: Chessboard, transpositions: ptr TTable, stopFlag, searchFlag: ptr Atomic[bool],
                       history: ptr HistoryTable, killers: ptr KillersTable): SearchManager =
    result = SearchManager(board: board, transpositionTable: transpositions, stopFlag: stopFlag,
                           searchFlag: searchFlag, history: history, killers: killers)
    for i in 0..MAX_DEPTH:
        for j in 0..MAX_DEPTH:
            result.pvMoves[i][j] = nullMove()



proc isSearching*(self: SearchManager): bool =
    ## Returns whether a search for the best
    ## move is in progress
    result = self.searchFlag[].load()


proc stop*(self: SearchManager) =
    ## Stops the search if it is
    ## running
    if self.isSearching():
        self.stopFlag[].store(true)


proc isKillerMove(self: SearchManager, move: Move, ply: int): bool =
    ## Returns whether the given move is a killer move
    when not defined(killers):
        return false
    else:
        for killer in self.killers[][ply]:
            if killer == move:
                return true


proc getEstimatedMoveScore(self: SearchManager, move: Move, ply: int): Score =
    ## Returns an estimated static score for the move used
    ## during move ordering
    result = Score(0)
    let 
        sideToMove = self.board.position.sideToMove
        nonSideToMove = sideToMove.opposite()

    let query = self.transpositionTable[].get(self.board.position.zobristKey)
    if query.success and query.entry.bestMove != nullMove() and query.entry.bestMove == move:
        return highestEval() + 1

    if ply > 0 and self.isKillerMove(move, ply):
        result += self.board.position.getPieceScore(move.startSquare) * 5

    if not move.isCapture():
        # History euristic bonus
        result += self.history[][sideToMove][move.startSquare][move.targetSquare]

    if move.isCapture():
        # Implementation of MVVLVA: Most Valuable Victim Least Valuable Attacker
        # We prioritize moves that capture the most valuable pieces, and as a
        # second goal we want to use our least valuable pieces to do so (this
        # is why we multiply the score of the captured piece by 10, to give
        # it priority)
        result += 10 * self.board.position.getPieceScore(move.targetSquare) - self.board.position.getPieceScore(move.startSquare)

    if move.isPromotion():
        # Promotions are a good idea to search first
        var piece: Piece
        case move.getPromotionType():
            of PromoteToBishop:
                piece = Piece(kind: Bishop, color: sideToMove)
            of PromoteToKnight:
                piece = Piece(kind: Knight, color: sideToMove)
            of PromoteToRook:
                piece = Piece(kind: Rook, color: sideToMove)
            of PromoteToQueen:
                piece = Piece(kind: Queen, color: sideToMove)
            else:
                discard  # Unreachable
        result += self.board.position.getPieceScore(piece, move.targetSquare)

    if self.board.position.getPawnAttacks(move.targetSquare, nonSideToMove) != 0:
        # Moving on a square attacked by an enemy pawn is _usually_ a very bad
        # idea. Assume the piece is lost and give a malus based on the fact that
        # losing a piece this way is dumb
        result -= self.board.position.getPieceScore(move.startSquare) * 2


proc reorderMoves(self: SearchManager, moves: var MoveList, ply: int) =
    ## Reorders the list of moves in-place, trying
    ## to place the best ones first
    
    proc orderer(a, b: Move): int {.closure.} =
        return cmp(self.getEstimatedMoveScore(a, ply), self.getEstimatedMoveScore(b, ply))
    
    # Ignore null moves beyond the lenght of the movelist
    sort(moves.data.toOpenArray(0, moves.len - 1), orderer, SortOrder.Descending)


proc timedOut(self: SearchManager): bool = getMonoTime() >= self.hardLimit
proc cancelled(self: SearchManager): bool = self.stopFlag[].load()
proc elapsedTime(self: SearchManager): int64 = (getMonoTime() - self.searchStart).inMilliseconds()


proc log(self: SearchManager, depth: int) =
    let 
        elapsedMsec = self.elapsedTime().uint64
        nps = 1000 * (self.nodeCount div max(elapsedMsec, 1))
    var logMsg = &"info depth {depth} seldepth {self.selectiveDepth} time {elapsedMsec} nodes {self.nodeCount} nps {nps}"
    logMsg &= &" hashfull {self.transpositionTable[].getFillEstimate()} score cp {self.bestRootScore}"
    if self.pvMoves[0][0] != nullMove():
        logMsg &= " pv "
        for move in self.pvMoves[0]:
            if move == nullMove():
                break
            logMsg &= &"{move.toAlgebraic()} "
    echo logMsg


proc shouldStop(self: SearchManager): bool =
    ## Returns whether searching should
    ## stop
    if self.cancelled():
        # Search has been cancelled!
        return true
    if self.timedOut():
        # We ran out of time!
        return true
    if self.maxNodes > 0 and self.nodeCount >= self.maxNodes:
        # Ran out of nodes
        return true


proc getSearchExtension(self: SearchManager, move: Move): int {.used.} =
    ## Returns the number of extensions that should be performed
    ## when exploring the given move
    if self.board.inCheck():
        # Opponent is in check: extend the search to see
        # if we can do other interesting things!
        return 1


proc getReduction(self: SearchManager, move: Move, depth, ply, moveNumber: int, isPV: bool): int =
    ## Returns the amount a search depth can be reduced to 
    
    # Move reduction gains: 54.1 +/- 25.5
    if moveNumber > 2 and depth > 2:
        result = LMR_TABLE[depth][moveNumber]
        if not isPV:
            # Reduce non PV nodes more
            inc(result)

        if self.board.inCheck():
            # Reduce less when opponent is in check
            dec(result)
        # Keep the reduction in the right range
        result = result.clamp(0, depth - 1)


proc qsearch(self: var SearchManager, ply: int, alpha, beta: Score): Score =
    ## Negamax search with a/b pruning that is restricted to
    ## capture moves (commonly called quiescent search). The
    ## purpose of this extra search step is to mitigate the 
    ## so called horizon effect that stems from the fact that,
    ## at some point, the engine will have to stop searching, possibly
    ## thinking a bad move is good because it couldn't see far enough
    ## ahead (this usually results in the engine blundering captures
    ## or sacking pieces for apparently no reason: the reason is that it
    ## did not look at the opponent's responses, because it stopped earlier. 
    ## That's the horizon). To address this, we look at all possible captures
    ## in the current position and make sure that a position is evaluated as 
    ## bad if only bad capture moves are possible, even if good non-capture moves
    ## exist
    if self.shouldStop():
        return
    if ply == MAX_DEPTH:
        return Score(0)
    let score = self.board.position.evaluate()
    if score >= beta:
        # Same as with the regular alpha-beta search
        return score
    if self.board.isDrawn():
        return Score(0)
    var moves = newMoveList()
    self.board.generateMoves(moves, capturesOnly=true)
    self.reorderMoves(moves, ply)
    var bestScore = score
    var alpha = max(alpha, score)
    for move in moves:
        self.board.doMove(move)
        inc(self.nodeCount)
        let score = -self.qsearch(ply + 1, -beta, -alpha)
        self.board.unmakeMove()
        bestScore = max(score, bestScore)
        if score >= beta:
            # This move was too good for us, opponent will not search it
            break
        # When a search is cancelled or times out, we need
        # to make sure the entire call stack unwinds back
        # to the root move. This is why the check is duplicated
        if self.shouldStop():
            return
        if score > alpha:
            alpha = score
    return bestScore


proc storeKillerMove(self: SearchManager, ply: int, move: Move) {.used.} =
    ## Stores a killer move into our killers table at the given
    ## ply

    # Stolen from https://rustic-chess.org/search/ordering/killers.html

    # First killer move must not be the same as the one we're storing
    let first = self.killers[][ply][0]
    if first == move:
        return
    var j = self.killers[][ply].len() - 2
    while j >= 0:
        # Shift moves one spot down
        self.killers[][ply][j + 1] = self.killers[][ply][j];
        dec(j)
    self.killers[][ply][0] = move


proc search(self: var SearchManager, depth, ply: int, alpha, beta: Score, isPV: bool): Score {.discardable.} =
    ## Negamax search with various optimizations and search features
    
    # Clear the PV table
    for i in 0..MAX_DEPTH:
        self.pvMoves[ply][i] = nullMove()
    if depth > 1 and self.shouldStop():
        # We do not let ourselves get cancelled at depth
        # one because then we wouldn't have a move to return.
        # In practice this should not be a problem
        return
    self.selectiveDepth = max(self.selectiveDepth, ply)
    if ply > 0:
        let query = self.transpositionTable[].get(self.board.position.zobristKey, depth.uint8)
        if query.success:
            case query.entry.flag:
                of Exact:
                    return query.entry.score
                of LowerBound:
                    if query.entry.score >= beta:
                        return query.entry.score
                of UpperBound:
                    if query.entry.score <= alpha:
                        return query.entry.score
    if self.board.isDrawn():
        return Score(0)
    if depth == 0:
        # Quiescent search gain: 264.8 +/- 71.6
        return self.qsearch(0, alpha, beta)
    var
        moves = newMoveList()
        depth = depth
        bestMove = nullMove()
    self.board.generateMoves(moves)
    self.reorderMoves(moves, ply)
    if moves.len() == 0:
        if self.board.inCheck():
            # Checkmate! We add the current ply
            # because mating in 3 is better than
            # mating in 5 (and conversely being
            # mated in 5 is better than being
            # mated in 3)
            return mateScore() + Score(ply)
        # Stalemate
        return Score(0)
    var bestScore = lowestEval()
    var alpha = alpha    
    let 
        sideToMove = self.board.position.sideToMove
        nonSideToMove {.used.} = sideToMove.opposite()
    for i, move in moves:
        if ply == 0 and self.searchMoves.len() > 0 and move notin self.searchMoves:
            continue
        self.board.doMove(move)
        let 
            extension = self.getSearchExtension(move)
            reduction = self.getReduction(move, depth, ply, i, isPV)
        inc(self.nodeCount)
        # Find the best move for us (worst move
        # for our opponent, hence the negative sign)
        var score: Score
        # Implementation of Principal Variation Search (PVS)
        if i == 0:
            # Due to our move ordering scheme, the first move is always the "best", so
            # search it always at full depth with the full search window
            score = -self.search(depth - 1 + extension, ply + 1, -beta, -alpha, isPV)
        elif reduction > 0:
            # Late Move Reductions: assume our move orderer did a good job,
            # so it is not worth it to look at all moves at the same depth equally.
            # If this move turns out to be better than we expected, we'll re-search
            # it at full depth

            # We first do a null-window search to see if there's a move that beats alpha
            # (we don't care about the actual value, so we search in the range [alpha, alpha + 1]
            # to increase the number of cutoffs)
            score = -self.search(depth - 1 - reduction, ply + 1, -alpha - 1, -alpha, isPV=false)
            # If the null window search beats alpha, we do a full window reduced search to get a
            # better feel for the actual score of the position. If the score turns out to beat alpha
            # (but not beta) again, we'll re-search this at full depth later
            if score > alpha:
                score = -self.search(depth - 1 - reduction, ply + 1, -beta, -alpha, isPV=false)
        else:
            # Move wasn't reduced, just do a null window search
            score = -self.search(depth - 1 + extension, ply + 1, -alpha - 1, -alpha, isPV=false)
        if i > 0 and score > alpha and score < beta:
            # The position failed high (and not low, which would mean it was too good for us and
            # our opponent wouldn't let us play it) in the null window search, search it
            # again with the full depth and full window
            score = -self.search(depth - 1 + extension, ply + 1, -beta, -alpha, isPV)
        self.board.unmakeMove()
        # When a search is cancelled or times out, we need
        # to make sure the entire call stack unwinds back
        # to the root move. This is why the check is duplicated
        if depth > 1 and self.shouldStop():
            return
        bestScore = max(score, bestScore)
        if score >= beta:
            if not move.isCapture() and self.history[][sideToMove][move.startSquare][move.targetSquare] < highestEval():
                # History euristic: keep track of moves that caused a beta cutoff and order
                # them early in subsequent searches, as they might be really good later. A
                # quadratic bonus wrt. depth is usually the value that is used (though some
                # engines, namely Stockfish, use a linear bonus. Maybe we can investigate this)
                self.history[][sideToMove][move.startSquare][move.targetSquare] += Score(depth * depth)
                # Killer move heuristic: store moves that caused a beta cutoff according to the distance from
                # root that they occurred at, as they might be good refutations for future moves from the opponent.
                when defined(killers):
                    self.storeKillerMove(ply, move)
            # This move was too good for us, opponent will not search it
            break
        if score > alpha:
            alpha = score
            bestMove = move
            if ply == 0:
                self.bestRootScore = score
            if isPV:
                # This loop is why pvMoves has one extra move.
                # We can just do ply + 1 and i + 1 without ever
                # fearing about buffer overflows
                for i, pv in self.pvMoves[ply + 1]:
                    if pv == nullMove():
                        self.pvMoves[ply][i + 1] = nullMove()
                        break
                    self.pvMoves[ply][i + 1] = pv
                self.pvMoves[ply][0] = move
        # TODO
        # else:
        #     when defined(noScaleHistory):
        #         if not move.isCapture() and self.history[][sideToMove][move.startSquare][move.targetSquare] > lowestEval():
        #             # Here, we punish moves that failed to raise alpha. This allows us to avoid scaling our values 
        #             # after every search (which should retain more information about the explored subtreees) and 
        #             # makes sure that moves that we thought were good but aren't are pushed further in the move list
        #             self.history[][sideToMove][move.startSquare][move.targetSquare] -= Score(depth * depth)
        #     else:
        #         discard
    let nodeType = if bestScore >= beta: LowerBound elif bestScore <= alpha: UpperBound else: Exact
    self.transpositionTable[].store(depth.uint8, bestScore, self.board.position.zobristKey, bestMove, nodeType)

    return bestScore


proc findBestMove*(self: var SearchManager, timeRemaining, increment: int64, maxDepth: int, maxNodes: uint64, searchMoves: seq[Move]): Move =
    ## Finds the best move in the current position
    ## and returns it, limiting search time according
    ## to the remaining time and increment values provided
    ## (in milliseconds) and only up to maxDepth ply (if maxDepth 
    ## is -1, a reasonable limit is picked). If maxNodes is supplied
    ## and is nonzero, search will stop once it has analyzed maxNodes
    ## nodes. If searchMoves is provided and is not empty, search will
    ## be restricted to the moves in the list. Note that regardless of
    ## any time limitations or explicit cancellations, the search will
    ## not stop until it  has at least cleared depth one. Search depth
    ## is always constrained to at most MAX_DEPTH ply from the root
    
    # Apparently negative remaining time is a thing. Welp
    let 
        maxSearchTime = max(1, (timeRemaining div 10) + (increment div 2))
        softLimit = maxSearchTime div 3
    result = nullMove()
    self.maxNodes = maxNodes
    self.searchMoves = searchMoves
    self.searchStart = getMonoTime()
    self.hardLimit = self.searchStart + initDuration(milliseconds=maxSearchTime)
    self.softLimit = self.searchStart + initDuration(milliseconds=softLimit)
    var maxDepth = maxDepth
    if maxDepth == -1:
        maxDepth = 30
    self.searchFlag[].store(true)
    # Iterative deepening loop
    for i in 1..min(MAX_DEPTH, maxDepth):
        self.search(i, 0, lowestEval(), highestEval(), true)
        if self.pvMoves[0][0] != nullMove():
            result = self.pvMoves[0][0]
        if self.shouldStop():
            self.log(i - 1)
            break
        self.log(i)
        # Soft time management: don't start a new search iteration
        # if the soft limit has expired, as it is unlikely to complete
        # anyway
        if getMonoTime() >= self.softLimit:
            break
    self.searchFlag[].store(false)
    self.stopFlag[].store(false)
